import abc
import typing as ta


T = ta.TypeVar('T')
U = ta.TypeVar('U')


class ValueNotPresentException(BaseException):
    pass


class Maybe(abc.ABC, ta.Generic[T]):

    @property
    @abc.abstractmethod
    def present(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def must(self) -> T:
        raise NotImplementedError

    @abc.abstractmethod
    def __call__(self) -> T:
        raise NotImplementedError

    @abc.abstractmethod
    def __iter__(self) -> ta.Iterator[T]:
        raise NotImplementedError

    @abc.abstractmethod
    def __getitem__(self, item: int) -> T:
        raise NotImplementedError

    @abc.abstractmethod
    def if_present(self, consumer: ta.Callable[[T], None]) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def filter(self, predicate: ta.Callable[[T], bool]) -> 'Maybe[T]':
        raise NotImplementedError

    @abc.abstractmethod
    def map(self, mapper: ta.Callable[[T], U]) -> 'Maybe[U]':
        raise NotImplementedError

    @abc.abstractmethod
    def flat_map(self, mapper: ta.Callable[[T], 'Maybe[U]']) -> 'Maybe[U]':
        raise NotImplementedError

    @abc.abstractmethod
    def or_else(self, other: T) -> 'Maybe[T]':
        raise NotImplementedError

    @abc.abstractmethod
    def or_else_get(self, supplier: ta.Callable[[], T]) -> 'Maybe[T]':
        raise NotImplementedError

    @abc.abstractmethod
    def or_else_raise(self, exception_supplier: ta.Callable[[], Exception]) -> 'Maybe[T]':
        raise NotImplementedError


class _Maybe(Maybe[T], tuple):
    __slots__ = ()

    def __init_subclass__(cls, **kwargs):
        raise TypeError

    @property
    def present(self) -> bool:
        return bool(self)

    def must(self) -> T:
        if not self:
            raise ValueNotPresentException
        return self[0]

    __call__ = must

    def __iter__(self) -> ta.Iterator[T]:
        raise TypeError

    locals()['__iter__'] = tuple.__iter__

    def __getitem__(self, item: int) -> T:  # type: ignore
        raise TypeError

    locals()['__getitem__'] = tuple.__getitem__

    def if_present(self, consumer: ta.Callable[[T], None]) -> None:
        if self:
            consumer(self[0])

    def filter(self, predicate: ta.Callable[[T], bool]) -> Maybe[T]:
        return self if self and predicate(self[0]) else _empty

    def map(self, mapper: ta.Callable[[T], U]) -> Maybe[U]:
        if self:
            value = mapper(self[0])
            if value is not None:
                return just(value)
        return _empty  # noqa

    def flat_map(self, mapper: ta.Callable[[T], Maybe[U]]) -> Maybe[U]:
        if self:
            value = mapper(self[0])
            if not isinstance(value, Maybe):
                raise TypeError(value)
            return value
        return _empty  # noqa

    def or_else(self, other: T) -> Maybe[T]:
        return self if self else just(other)

    def or_else_get(self, supplier: ta.Callable[[], T]) -> Maybe[T]:
        return self if self else just(supplier())

    def or_else_raise(self, exception_supplier: ta.Callable[[], Exception]) -> Maybe[T]:
        if self:
            return self
        raise exception_supplier()


def just(v: T) -> Maybe[T]:
    return tuple.__new__(_Maybe, (v,))  # noqa


_empty = tuple.__new__(_Maybe, ())


def empty() -> Maybe[T]:
    return _empty  # noqa


def maybe(o: T | None) -> Maybe[T]:
    if o is None:
        return _empty  # noqa
    return just(o)



from .descriptor import Ignore  # noqa
from .descriptor import Scope  # noqa
from .descriptor import cache  # noqa
from .descriptor import ignore  # noqa
from .impl import LFU  # noqa
from .impl import LRI  # noqa
from .impl import LRU  # noqa
from .impl import new_cache  # noqa
from .types import Cache  # noqa
from .types import Eviction  # noqa
from .types import OverweightError  # noqa



import abc
import time
import typing as ta


TimeoutLike: ta.TypeAlias = ta.Union['Timeout', float]


class Timeout(abc.ABC):
    @abc.abstractmethod
    def __call__(self) -> float:
        raise NotImplementedError

    @abc.abstractmethod
    def or_(self, o: ta.Any) -> ta.Any:
        raise NotImplementedError


class DeadlineTimeout(Timeout):
    def __init__(
            self,
            deadline: float,
            exc: type[BaseException] | BaseException = TimeoutError,
    ) -> None:
        super().__init__()
        self.deadline = deadline
        self.exc = exc

    def __call__(self) -> float:
        if (rem := self.deadline - time.time()) > 0:
            return rem
        raise self.exc

    def or_(self, o: ta.Any) -> ta.Any:
        return self()


class InfiniteTimeout(Timeout):
    def __call__(self) -> float:
        return float('inf')

    def or_(self, o: ta.Any) -> ta.Any:
        return o


def timeout(t: TimeoutLike | None) -> Timeout:
    if t is None:
        return InfiniteTimeout()
    if isinstance(t, Timeout):
        return t
    if isinstance(t, (float, int)):
        return DeadlineTimeout(time.time() + t)
    raise TypeError(t)



from ....diag import pydevd as opd
from ._registry import register


@register
class PydevdPlugin:

    def pytest_collection(self, session):
        setup = opd.get_setup()
        if setup is not None:
            if hasattr(session.config, '_env_timeout'):
                session.config._env_timeout = None  # noqa



import contextlib
import functools
import importlib.util
import sys
import types
import typing as ta

from .cached import cached_function


##


def can_import(name: str, package: str | None = None) -> bool:
    return importlib.util.find_spec(name, package) is not None


##


def lazy_import(name: str, package: str | None = None) -> ta.Callable[[], ta.Any]:
    return cached_function(functools.partial(importlib.import_module, name, package=package))


def proxy_import(name: str, package: str | None = None) -> types.ModuleType:
    omod = None

    def __getattr__(att):  # noqa
        nonlocal omod
        if omod is None:
            omod = importlib.import_module(name, package=package)
        return getattr(omod, att)

    lmod = types.ModuleType(name)
    lmod.__getattr__ = __getattr__  # type: ignore
    return lmod


##


def import_module(dotted_path: str) -> types.ModuleType:
    if not dotted_path:
        raise ImportError(dotted_path)
    mod = __import__(dotted_path, globals(), locals(), [])
    for name in dotted_path.split('.')[1:]:
        try:
            mod = getattr(mod, name)
        except AttributeError:
            raise AttributeError(f'Module {mod!r} has no attribute {name!r}') from None
    return mod


def import_module_attr(dotted_path: str) -> ta.Any:
    module_name, _, class_name = dotted_path.rpartition('.')
    mod = import_module(module_name)
    try:
        return getattr(mod, class_name)
    except AttributeError:
        raise AttributeError(f'Module {module_name!r} has no attr {class_name!r}') from None


SPECIAL_IMPORTABLE: ta.AbstractSet[str] = frozenset([
    '__init__.py',
    '__main__.py',
])


def yield_importable(
        package_root: str,
        *,
        recursive: bool = False,
        filter: ta.Callable[[str], bool] | None = None,  # noqa
        include_special: bool = False,
) -> ta.Iterator[str]:
    from importlib import resources

    def rec(cur):
        if cur.split('.')[-1] == '__pycache__':
            return

        try:
            module = sys.modules[cur]
        except KeyError:
            try:
                __import__(cur)
            except ImportError:
                return
            module = sys.modules[cur]

        # FIXME: pyox
        if getattr(module, '__file__', None) is None:
            return

        for file in resources.files(cur).iterdir():
            if file.is_file() and file.name.endswith('.py'):
                if not (include_special or file.name not in SPECIAL_IMPORTABLE):
                    continue

                name = cur + '.' + file.name[:-3]
                if filter is not None and not filter(name):
                    continue

                yield name

            elif recursive and file.is_dir():
                name = cur + '.' + file.name
                if filter is not None and not filter(name):
                    continue
                with contextlib.suppress(ImportError, NotImplementedError):
                    yield from rec(name)

    yield from rec(package_root)


def yield_import_all(
        package_root: str,
        *,
        globals: dict[str, ta.Any] | None = None,  # noqa
        locals: dict[str, ta.Any] | None = None,  # noqa
        recursive: bool = False,
        filter: ta.Callable[[str], bool] | None = None,  # noqa
        include_special: bool = False,
) -> ta.Iterator[str]:
    for import_path in yield_importable(
            package_root,
            recursive=recursive,
            filter=filter,
            include_special=include_special,
    ):
        __import__(import_path, globals=globals, locals=locals)
        yield import_path


def import_all(
        package_root: str,
        *,
        recursive: bool = False,
        filter: ta.Callable[[str], bool] | None = None,  # noqa
        include_special: bool = False,
) -> None:
    for _ in yield_import_all(
            package_root,
            recursive=recursive,
            filter=filter,
            include_special=include_special,
    ):
        pass


def try_import(spec: str) -> types.ModuleType | None:
    s = spec.lstrip('.')
    l = len(spec) - len(s)
    try:
        return __import__(s, globals(), level=l)
    except ImportError:
        return None


##


def resolve_import_name(name: str, package: str | None = None) -> str:
    level = 0

    if name.startswith('.'):
        if not package:
            raise TypeError("the 'package' argument is required to perform a relative import for {name!r}")
        for character in name:
            if character != '.':
                break
            level += 1

    name = name[level:]

    if not isinstance(name, str):
        raise TypeError(f'module name must be str, not {type(name)}')
    if level < 0:
        raise ValueError('level must be >= 0')
    if level > 0:
        if not isinstance(package, str):
            raise TypeError('__package__ not set to a string')
        elif not package:
            raise ImportError('attempted relative import with no known parent package')
    if not name and level == 0:
        raise ValueError('Empty module name')

    if level > 0:
        bits = package.rsplit('.', level - 1)  # type: ignore
        if len(bits) < level:
            raise ImportError('attempted relative import beyond top-level package')
        base = bits[0]
        name = f'{base}.{name}' if name else base

    return name


##


_REGISTERED_CONDITIONAL_IMPORTS: dict[str, list[str] | None] = {}


def _register_conditional_import(when: str, then: str, package: str | None = None) -> None:
    wn = resolve_import_name(when, package)
    tn = resolve_import_name(then, package)
    if tn in sys.modules:
        return
    if wn in sys.modules:
        __import__(tn)
    else:
        tns = _REGISTERED_CONDITIONAL_IMPORTS.setdefault(wn, [])
        if tns is None:
            raise Exception(f'Conditional import trigger already cleared: {wn=} {tn=}')
        tns.append(tn)


def _trigger_conditional_imports(package: str) -> None:
    tns = _REGISTERED_CONDITIONAL_IMPORTS.get(package, [])
    if tns is None:
        raise Exception(f'Conditional import trigger already cleared: {package=}')
    _REGISTERED_CONDITIONAL_IMPORTS[package] = None
    for tn in tns:
        __import__(tn)



import typing as ta


DBAPITypeCode: ta.TypeAlias = ta.Any | None

DBAPIColumnDescription: ta.TypeAlias = tuple[
    str,
    DBAPITypeCode,
    int | None,
    int | None,
    int | None,
    int | None,
    bool | None,
]


class DBAPIConnection(ta.Protocol):
    def close(self) -> object: ...

    def commit(self) -> object: ...

    # optional:
    # def rollback(self) -> ta.Any: ...

    def cursor(self) -> 'DBAPICursor': ...


class DBAPICursor(ta.Protocol):
    @property
    def description(self) -> ta.Sequence[DBAPIColumnDescription] | None: ...

    @property
    def rowcount(self) -> int: ...

    # optional:
    # def callproc(self, procname: str, parameters: Sequence[ta.Any] = ...) -> Sequence[ta.Any]: ...

    def close(self) -> object: ...

    def execute(
        self,
        operation: str,
        parameters: ta.Sequence[ta.Any] | ta.Mapping[str, ta.Any] = ...,
    ) -> object: ...

    def executemany(
        self,
        operation: str,
        seq_of_parameters: ta.Sequence[ta.Sequence[ta.Any]],
    ) -> object: ...

    def fetchone(self) -> ta.Sequence[ta.Any] | None: ...

    def fetchmany(self, size: int = ...) -> ta.Sequence[ta.Sequence[ta.Any]]: ...

    def fetchall(self) -> ta.Sequence[ta.Sequence[ta.Any]]: ...

    # optional:
    # def nextset(self) -> None | Literal[True]: ...

    arraysize: int

    def setinputsizes(self, sizes: ta.Sequence[DBAPITypeCode | int | None]) -> object: ...

    def setoutputsize(self, size: int, column: int = ...) -> object: ...



"""
def loads(obj: str | bytes | bytearray | memoryview) -> ta.Any | oj.JSONDEcodeError
def dumps(obj: ta.Any, **DumpOpts) -> bytes
"""
import dataclasses as dc
import typing as ta

from .... import lang


if ta.TYPE_CHECKING:
    import orjson as oj
else:
    oj = lang.proxy_import('orjson')


@dc.dataclass(frozen=True, kw_only=True)
class Options:
    append_newline: bool = False  # append \n to the output

    indent_2: bool = False  # pretty-print output with an indent of two spaces.

    naive_utc: bool = False  # serialize datetime.datetime objects without a tzinfo as UTC

    non_str_keys: bool = False  # serialize dict keys of type other than str

    omit_microseconds: bool = False  # do not serialize the microsecond field on datetime.datetime and datetime.time instances  # noqa

    passthrough_dataclass: bool = False  # passthrough dataclasses.dataclass instances to default
    passthrough_datetime: bool = False  # passthrough datetime.datetime, datetime.date, and datetime.time instances to default  # noqa
    passthrough_subclass: bool = False  # passthrough subclasses of builtin types to default

    serialize_numpy: bool = False  # serialize numpy.ndarray instances

    sort_keys: bool = False  # serialize dict keys in sorted order - the default is to serialize in an unspecified order

    strict_integer: bool = False  # enforce 53-bit limit on integers

    utc_z: bool = False  # serialize a UTC timezone on datetime.datetime instances as Z instead of +00:00

    ##

    def __int__(self) -> int:
        return (
            (oj.OPT_APPEND_NEWLINE if self.append_newline else 0) |

            (oj.OPT_INDENT_2 if self.indent_2 else 0) |

            (oj.OPT_NAIVE_UTC if self.naive_utc else 0) |

            (oj.OPT_NON_STR_KEYS if self.non_str_keys else 0) |

            (oj.OPT_OMIT_MICROSECONDS if self.omit_microseconds else 0) |

            (oj.OPT_PASSTHROUGH_DATACLASS if self.passthrough_dataclass else 0) |
            (oj.OPT_PASSTHROUGH_DATETIME if self.passthrough_datetime else 0) |
            (oj.OPT_PASSTHROUGH_SUBCLASS if self.passthrough_subclass else 0) |

            (oj.OPT_SERIALIZE_NUMPY if self.serialize_numpy else 0) |

            (oj.OPT_SORT_KEYS if self.sort_keys else 0) |

            (oj.OPT_STRICT_INTEGER if self.strict_integer else 0) |

            (oj.OPT_UTC_Z if self.utc_z else 0)
        )


@dc.dataclass(frozen=True, kw_only=True)
class DumpOpts:
    default: ta.Callable[[ta.Any], ta.Any] | None = None
    option: Options = Options()



"""
TODO:
 - default command
 - auto match all underscores to hyphens
"""
import argparse
import dataclasses as dc
import functools
import sys
import typing as ta

from . import c3
from . import check


T = ta.TypeVar('T')


##


SUPPRESS = argparse.SUPPRESS

OPTIONAL = argparse.OPTIONAL
ZERO_OR_MORE = argparse.ZERO_OR_MORE
ONE_OR_MORE = argparse.ONE_OR_MORE
PARSER = argparse.PARSER
REMAINDER = argparse.REMAINDER

HelpFormatter = argparse.HelpFormatter
RawDescriptionHelpFormatter = argparse.RawDescriptionHelpFormatter
RawTextHelpFormatter = argparse.RawTextHelpFormatter
ArgumentDefaultsHelpFormatter = argparse.ArgumentDefaultsHelpFormatter

MetavarTypeHelpFormatter = argparse.MetavarTypeHelpFormatter

ArgumentError = argparse.ArgumentError
ArgumentTypeError = argparse.ArgumentTypeError

Action = argparse.Action
BooleanOptionalAction = argparse.BooleanOptionalAction
SubParsersAction = argparse._SubParsersAction  # noqa

FileType = argparse.FileType

Namespace = argparse.Namespace

ArgumentParser = argparse.ArgumentParser


##


@dc.dataclass(eq=False)
class Arg:
    args: ta.Sequence[ta.Any]
    kwargs: ta.Mapping[str, ta.Any]
    dest: str | None = None

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return getattr(instance.args, self.dest)  # type: ignore


def arg(*args, **kwargs) -> Arg:
    return Arg(args, kwargs)


#


CommandFn = ta.Callable[[], None]


@dc.dataclass(eq=False)
class Command:
    name: str
    fn: CommandFn
    args: ta.Sequence[Arg] = ()
    parent: ta.Optional['Command'] = None
    accepts_unknown: bool = False

    def __post_init__(self) -> None:
        check.isinstance(self.name, str)
        check.not_in('_', self.name)
        check.not_empty(self.name)

        check.callable(self.fn)
        check.arg(all(isinstance(a, Arg) for a in self.args))
        check.isinstance(self.parent, (Command, None))
        check.isinstance(self.accepts_unknown, bool)

        functools.update_wrapper(self, self.fn)

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return dc.replace(self, fn=self.fn.__get__(instance, owner))

    def __call__(self, *args, **kwargs) -> None:
        return self.fn(*args, **kwargs)


def command(
        *args: Arg,
        name: str | None = None,
        parent: Command | None = None,
        accepts_unknown: bool = False,
) -> ta.Any:  # ta.Callable[[CommandFn], Command]:  # FIXME
    for arg in args:
        check.isinstance(arg, Arg)
    check.isinstance(name, (str, None))
    check.isinstance(parent, (Command, None))

    def inner(fn):
        return Command(
            (name if name is not None else fn.__name__).replace('_', '-'),
            fn,
            args,
            parent=parent,
            accepts_unknown=accepts_unknown,
        )

    return inner


##


def get_arg_ann_kwargs(ann: ta.Any) -> ta.Mapping[str, ta.Any]:
    if ann is str:
        return {}
    elif ann is int:
        return {'type': int}
    elif ann is bool:
        return {'action': 'store_true'}
    elif ann is list:
        return {'action': 'append'}
    else:
        raise TypeError(ann)


class _AnnotationBox:

    def __init__(self, annotations: ta.Mapping[str, ta.Any]) -> None:
        super().__init__()
        self.__annotations__ = annotations  # type: ignore


class _CliMeta(type):

    def __new__(mcls, name: str, bases: ta.Sequence[type], namespace: ta.Mapping[str, ta.Any]) -> type:
        if not bases:
            return super().__new__(mcls, name, tuple(bases), dict(namespace))

        bases = list(bases)
        namespace = dict(namespace)

        objs = {}
        mro = c3.merge([list(b.__mro__) for b in bases])
        for bns in [bcls.__dict__ for bcls in reversed(mro)] + [namespace]:
            bseen = set()  # type: ignore
            for k, v in bns.items():
                if isinstance(v, (Command, Arg)):
                    check.not_in(v, bseen)
                    bseen.add(v)
                    objs[k] = v
                elif k in objs:
                    del [k]

        anns = ta.get_type_hints(_AnnotationBox({
            **{k: v for bcls in reversed(mro) for k, v in getattr(bcls, '__annotations__', {}).items()},
            **namespace.get('__annotations__', {}),
        }), globalns=namespace.get('__globals__', {}))

        if 'parser' in namespace:
            parser = check.isinstance(namespace.pop('parser'), ArgumentParser)
        else:
            parser = ArgumentParser()
        namespace['_parser'] = parser

        subparsers = parser.add_subparsers()
        for att, obj in objs.items():
            if isinstance(obj, Command):
                if obj.parent is not None:
                    raise NotImplementedError
                cparser = subparsers.add_parser(obj.name)
                for arg in (obj.args or []):
                    cparser.add_argument(*arg.args, **arg.kwargs)
                cparser.set_defaults(_cmd=obj)

            elif isinstance(obj, Arg):
                if att in anns:
                    akwargs = get_arg_ann_kwargs(anns[att])
                    obj.kwargs = {**akwargs, **obj.kwargs}
                if not obj.dest:
                    if 'dest' in obj.kwargs:
                        obj.dest = obj.kwargs['dest']
                    else:
                        obj.dest = obj.kwargs['dest'] = att  # type: ignore
                parser.add_argument(*obj.args, **obj.kwargs)

            else:
                raise TypeError(obj)

        return super().__new__(mcls, name, tuple(bases), namespace)


class Cli(metaclass=_CliMeta):

    def __init__(self, argv: ta.Sequence[str] | None = None) -> None:
        super().__init__()

        self._argv = argv if argv is not None else sys.argv[1:]

        self._args, self._unknown_args = self.get_parser().parse_known_args(self._argv)

    _parser: ta.ClassVar[ArgumentParser]

    @classmethod
    def get_parser(cls) -> ArgumentParser:
        return cls._parser

    @property
    def argv(self) -> ta.Sequence[str]:
        return self._argv

    @property
    def args(self) -> Namespace:
        return self._args

    @property
    def unknown_args(self) -> ta.Sequence[str]:
        return self._unknown_args

    def _run_cmd(self, cmd: Command) -> None:
        cmd.__get__(self, type(self))()

    def __call__(self) -> None:
        cmd = getattr(self.args, '_cmd', None)

        if self._unknown_args and not (cmd is not None and cmd.accepts_unknown):
            msg = f'unrecognized arguments: {" ".join(self._unknown_args)}'
            if (parser := self.get_parser()).exit_on_error:  # type: ignore
                parser.error(msg)
            else:
                raise ArgumentError(None, msg)

        if cmd is None:
            self.get_parser().print_help()
            return

        self._run_cmd(cmd)



"""
TODO:
 - import mimetypes lol
"""
import base64
import http


##


def format_status(status: http.HTTPStatus) -> str:
    return '%d %s' % (int(status), status.phrase)


STATUS_OK = format_status(http.HTTPStatus.OK)

STATUS_FOUND = format_status(http.HTTPStatus.FOUND)
STATUS_TEMPORARY_REDIRECT = format_status(http.HTTPStatus.TEMPORARY_REDIRECT)

STATUS_BAD_REQUEST = format_status(http.HTTPStatus.BAD_REQUEST)
STATUS_UNAUTHORIZED = format_status(http.HTTPStatus.UNAUTHORIZED)
STATUS_FORBIDDEN = format_status(http.HTTPStatus.FORBIDDEN)
STATUS_NOT_FOUND = format_status(http.HTTPStatus.NOT_FOUND)
STATUS_METHOD_NOT_ALLOWED = format_status(http.HTTPStatus.METHOD_NOT_ALLOWED)
STATUS_REQUEST_TIMEOUT = format_status(http.HTTPStatus.REQUEST_TIMEOUT)

STATUS_INTERNAL_SERVER_ERROR = format_status(http.HTTPStatus.INTERNAL_SERVER_ERROR)
STATUS_NOT_IMPLEMENTED = format_status(http.HTTPStatus.NOT_IMPLEMENTED)
STATUS_BAD_GATEWAY = format_status(http.HTTPStatus.BAD_GATEWAY)
STATUS_SERVICE_UNAVAILABLE = format_status(http.HTTPStatus.SERVICE_UNAVAILABLE)
STATUS_GATEWAY_TIMEOUT = format_status(http.HTTPStatus.GATEWAY_TIMEOUT)


##


HEADER_CONTENT_TYPE = b'Content-Type'
HEADER_ACCEPT = b'Accept'

CONTENT_CHARSET_UTF8 = b'charset=utf-8'

CONTENT_TYPE_BYTES = b'application/octet-stream'

CONTENT_TYPE_HTML = b'text/html'
CONTENT_TYPE_HTML_UTF8 = b'; '.join([CONTENT_TYPE_HTML, CONTENT_CHARSET_UTF8])

CONTENT_TYPE_ICON = b'image/x-icon'

CONTENT_TYPE_JSON = b'application/json'
CONTENT_TYPE_JSON_UTF8 = b'; '.join([CONTENT_TYPE_JSON, CONTENT_CHARSET_UTF8])

CONTENT_TYPE_PNG = b'image/png'

CONTENT_TYPE_TEXT = b'text/plain'
CONTENT_TYPE_TEXT_UTF8 = b'; '.join([CONTENT_TYPE_TEXT, CONTENT_CHARSET_UTF8])


##


HEADER_AUTH = b'Authorization'
BEARER_AUTH_HEADER_PREFIX = b'Bearer '
BASIC_AUTH_HEADER_PREFIX = b'Basic '


def format_basic_auth_header(username: str, password: str) -> bytes:
    return BASIC_AUTH_HEADER_PREFIX + base64.b64encode(':'.join([username, password]).encode())



import dataclasses as dc
import os
import typing as ta

from .. import lang


if ta.TYPE_CHECKING:
    import psutil as _psutil
else:
    _psutil = lang.proxy_import('psutil')


##


@dc.dataclass(frozen=True)
class Times:
    user: float
    system: float
    children_user: float
    children_system: float
    elapsed: float


def times() -> Times:
    t = os.times()
    return Times(**{f.name: getattr(t, f.name) for f in dc.fields(Times)})


##


@dc.dataclass(frozen=True, kw_only=True)
class ProcStats:
    pid: int

    rss: int


def get_psutil_procstats(pid: int | None = None) -> ProcStats:
    if pid is None:
        pid = os.getpid()

    proc = _psutil.Process(pid)
    mi = proc.memory_info()

    return ProcStats(
        pid=pid,

        rss=mi.rss,
    )



from .json import (  # noqa
    COMPACT_KWARGS,
    COMPACT_SEPARATORS,
    PRETTY_INDENT,
    PRETTY_KWARGS,
    detect_encoding,
    dump,
    dump_compact,
    dump_pretty,
    dumps,
    dumps_compact,
    dumps_pretty,
    load,
    loads,
)

from .render import (  # noqa
    JsonRenderer,
)



import dataclasses as dc
import threading
import typing as ta

from .. import check
from .. import matchfns as mfs
from .. import reflect as rfl


R = ta.TypeVar('R')
C = ta.TypeVar('C')


##


@dc.dataclass(frozen=True)
class TypeMapFactory(mfs.MatchFn[[C, rfl.Type], R]):
    m: ta.Mapping[rfl.Type, R] = dc.field(default_factory=dict)

    def __post_init__(self) -> None:
        for k in self.m:
            if not isinstance(k, rfl.TYPES):
                raise TypeError(k)

    def guard(self, ctx: C, rty: rfl.Type) -> bool:
        check.isinstance(rty, rfl.TYPES)
        return rty in self.m

    def fn(self, ctx: C, rty: rfl.Type) -> R:
        check.isinstance(rty, rfl.TYPES)
        try:
            return self.m[rty]
        except KeyError:
            raise mfs.MatchGuardError(ctx, rty)  # noqa


##


class TypeCacheFactory(mfs.MatchFn[[C, rfl.Type], R]):
    def __init__(self, f: mfs.MatchFn[[C, rfl.Type], R]) -> None:
        super().__init__()
        self._f = f
        self._dct: dict[rfl.Type, R | None] = {}
        self._mtx = threading.RLock()

    def guard(self, ctx: C, rty: rfl.Type) -> bool:
        check.isinstance(rty, rfl.TYPES)
        with self._mtx:
            try:
                e = self._dct[rty]
            except KeyError:
                if self._f.guard(ctx, rty):
                    return True
                else:
                    self._dct[rty] = None
                    return False
            else:
                return e is not None

    def fn(self, ctx: C, rty: rfl.Type) -> R:
        check.isinstance(rty, rfl.TYPES)
        with self._mtx:
            try:
                e = self._dct[rty]
            except KeyError:
                try:
                    ret = self._f(ctx, rty)
                except mfs.MatchGuardError:
                    self._dct[rty] = None
                    raise
                else:
                    self._dct[rty] = ret
                    return ret
            else:
                if e is None:
                    raise mfs.MatchGuardError(ctx, rty)
                else:
                    return e


##


class RecursiveTypeFactory(mfs.MatchFn[[C, rfl.Type], R]):
    def __init__(
            self,
            f: mfs.MatchFn[[C, rfl.Type], R],
            prx: ta.Callable[[], tuple[R, ta.Callable[[R], None]]],
    ) -> None:
        super().__init__()
        self._f = f
        self._prx = prx
        self._dct: dict[rfl.Type, R] = {}

    def guard(self, ctx: C, rty: rfl.Type) -> bool:
        check.isinstance(rty, rfl.TYPES)
        return self._f.guard(ctx, rty)

    def fn(self, ctx: C, rty: rfl.Type) -> R:
        check.isinstance(rty, rfl.TYPES)
        try:
            return self._dct[rty]
        except KeyError:
            pass
        p, sp = self._prx()
        self._dct[rty] = p
        try:
            r = self._f(ctx, rty)
            sp(r)
            return r
        finally:
            del self._dct[rty]



#!/usr/bin/env python3
import argparse
import json
import pprint
import sys

from . import exceptions
from .parser import compile
from .parser import search


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('expression')
    parser.add_argument(
        '-f',
        '--filename',
        help='The filename containing the input data.  If a filename is not given then data is read from stdin.',
    )
    parser.add_argument(
        '--ast',
        action='store_true',
        help='Pretty print the AST, do not search the data.',
    )
    args = parser.parse_args()

    expression = args.expression
    if args.ast:
        # Only print the AST
        expression = compile(args.expression)
        sys.stdout.write(pprint.pformat(expression.parsed))
        sys.stdout.write('\n')
        return 0

    if args.filename:
        with open(args.filename) as f:
            data = json.load(f)
    else:
        data = sys.stdin.read()
        data = json.loads(data)

    try:
        sys.stdout.write(json.dumps(search(expression, data), indent=4, ensure_ascii=False))
        sys.stdout.write('\n')

    except exceptions.ArityError as e:
        sys.stderr.write(f'invalid-arity: {e}\n')
        return 1

    except exceptions.JmespathTypeError as e:
        sys.stderr.write(f'invalid-type: {e}\n')
        return 1

    except exceptions.JmespathValueError as e:
        sys.stderr.write(f'invalid-value: {e}\n')
        return 1

    except exceptions.UnknownFunctionError as e:
        sys.stderr.write(f'unknown-function: {e}\n')
        return 1

    except exceptions.ParseError as e:
        sys.stderr.write(f'syntax-error: {e}\n')
        return 1


if __name__ == '__main__':
    sys.exit(_main())



"""
TODO:
 - auto-gen from __subclasses__ if abstract
  - cfg: unless prefixed with _ or abstract
  - iff Sealed
 - auto-name
"""
import collections.abc
import dataclasses as dc
import typing as ta

from .. import check
from .. import lang
from .. import reflect as rfl
from .base import MarshalContext
from .base import Marshaler
from .base import MarshalerFactory
from .base import UnmarshalContext
from .base import Unmarshaler
from .base import UnmarshalerFactory
from .naming import Naming
from .naming import translate_name
from .registries import RegistryItem
from .values import Value


##


class TypeTagging(RegistryItem, lang.Abstract, lang.Sealed):
    pass


class WrapperTypeTagging(TypeTagging, lang.Final):
    pass


@dc.dataclass(frozen=True)
class FieldTypeTagging(TypeTagging, lang.Final):
    field: str


##


@dc.dataclass(frozen=True)
class Impl:
    ty: type
    tag: str
    alts: ta.AbstractSet[str] = frozenset()


class Polymorphism:
    def __init__(
            self,
            ty: type,
            impls: ta.Iterable[Impl],
    ) -> None:
        super().__init__()
        self._ty = ty
        self._impls = list(impls)

        by_ty: dict[type, Impl] = {}
        by_tag: dict[str, Impl] = {}
        for i in self._impls:
            if not issubclass(i.ty, ty) or i.ty in by_ty:
                raise TypeError(i.ty, ty)
            if i.tag in by_tag:
                raise NameError(i.tag)
            for a in i.alts:
                if a in by_tag:
                    raise NameError(a)
            by_ty[i.ty] = i
            by_tag[i.tag] = i
            for a in i.alts:
                by_tag[a] = i
        self._by_ty = by_ty
        self._by_tag = by_tag

    @property
    def ty(self) -> type:
        return self._ty

    @property
    def impls(self) -> ta.Sequence[Impl]:
        return self._impls

    @property
    def by_ty(self) -> ta.Mapping[type, Impl]:
        return self._by_ty

    @property
    def by_tag(self) -> ta.Mapping[str, Impl]:
        return self._by_tag


def polymorphism_from_subclasses(
        ty: type,
        *,
        naming: Naming | None = None,
        strip_suffix: bool = False,
) -> Polymorphism:
    dct: dict[str, Impl] = {}

    seen: set[type] = set()
    todo: list[type] = [ty]
    while todo:
        cur = todo.pop()
        seen.add(cur)

        todo.extend(nxt for nxt in cur.__subclasses__() if nxt not in seen)

        if lang.is_abstract_class(cur):
            continue

        name = cur.__name__
        if strip_suffix:
            name = lang.strip_suffix(name, ty.__name__)
        if naming is not None:
            name = translate_name(name, naming)
        if name in dct:
            raise KeyError(f'Duplicate name: {name}')

        dct[name] = Impl(
            cur,
            name,
        )

    return Polymorphism(ty, dct.values())


##


@dc.dataclass(frozen=True)
class WrapperPolymorphismMarshaler(Marshaler):
    m: ta.Mapping[type, tuple[str, Marshaler]]

    def marshal(self, ctx: MarshalContext, o: ta.Any | None) -> Value:
        tag, m = self.m[type(o)]
        return {tag: m.marshal(ctx, o)}


@dc.dataclass(frozen=True)
class FieldPolymorphismMarshaler(Marshaler):
    m: ta.Mapping[type, tuple[str, Marshaler]]
    tf: str

    def marshal(self, ctx: MarshalContext, o: ta.Any | None) -> Value:
        tag, m = self.m[type(o)]
        return {self.tf: tag, **m.marshal(ctx, o)}  # type: ignore


@dc.dataclass(frozen=True)
class PolymorphismMarshalerFactory(MarshalerFactory):
    p: Polymorphism
    tt: TypeTagging = WrapperTypeTagging()

    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return rty is self.p.ty

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        check.is_(rty, self.p.ty)
        m = {
            i.ty: (i.tag, ctx.make(i.ty))
            for i in self.p.impls
        }
        if isinstance(self.tt, WrapperTypeTagging):
            return WrapperPolymorphismMarshaler(m)
        elif isinstance(self.tt, FieldTypeTagging):
            return FieldPolymorphismMarshaler(m, self.tt.field)
        else:
            raise TypeError(self.tt)


##


@dc.dataclass(frozen=True)
class WrapperPolymorphismUnmarshaler(Unmarshaler):
    m: ta.Mapping[str, Unmarshaler]

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any | None:
        ma = check.isinstance(v, collections.abc.Mapping)
        [(tag, iv)] = ma.items()
        u = self.m[tag]  # type: ignore
        return u.unmarshal(ctx, iv)  # type: ignore


@dc.dataclass(frozen=True)
class FieldPolymorphismUnmarshaler(Unmarshaler):
    m: ta.Mapping[str, Unmarshaler]
    tf: str

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any | None:
        ma = dict(check.isinstance(v, collections.abc.Mapping))
        tag = ma.pop(self.tf)  # type: ignore
        u = self.m[tag]  # type: ignore
        return u.unmarshal(ctx, ma)


@dc.dataclass(frozen=True)
class PolymorphismUnmarshalerFactory(UnmarshalerFactory):
    p: Polymorphism
    tt: TypeTagging = WrapperTypeTagging()

    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return rty is self.p.ty

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        check.is_(rty, self.p.ty)
        m = {
            t: u
            for i in self.p.impls
            for u in [ctx.make(i.ty)]
            for t in [i.tag, *i.alts]
        }
        if isinstance(self.tt, WrapperTypeTagging):
            return WrapperPolymorphismUnmarshaler(m)
        elif isinstance(self.tt, FieldTypeTagging):
            return FieldPolymorphismUnmarshaler(m, self.tt.field)
        else:
            raise TypeError(self.tt)



import types
import typing as ta
import weakref

from .classes.abstract import is_abstract_class


T = ta.TypeVar('T')


##


def attr_repr(obj: ta.Any, *attrs: str) -> str:
    return f'{type(obj).__name__}({", ".join(f"{attr}={getattr(obj, attr)!r}" for attr in attrs)})'


def arg_repr(*args, **kwargs) -> str:
    return ', '.join(*(
        list(map(repr, args)) +
        [f'{k}={v!r}' for k, v in kwargs.items()]
    ))


def opt_repr(obj: ta.Any) -> str | None:
    if obj is None:
        return None
    return repr(obj)


##


_CAN_WEAKREF_TYPE_MAP: ta.MutableMapping[type, bool] = weakref.WeakKeyDictionary()


def can_weakref(obj: ta.Any) -> bool:
    _type = type(obj)
    try:
        return _CAN_WEAKREF_TYPE_MAP[_type]
    except KeyError:
        pass
    try:
        weakref.ref(obj)
    except TypeError:
        ret = False
    else:
        ret = True
    _CAN_WEAKREF_TYPE_MAP[_type] = ret
    return ret


##


def new_type(
        name: str,
        bases: ta.Sequence[ta.Any],
        namespace: ta.Mapping[str, ta.Any],
        **kwargs,
) -> type:
    return types.new_class(
        name,
        tuple(bases),
        kwds=kwargs,
        exec_body=lambda ns: ns.update(namespace),
    )


def super_meta(
        super_meta: ta.Any,
        meta: type,
        name: str,
        bases: ta.Sequence[ta.Any],
        namespace: ta.MutableMapping[str, ta.Any],
        **kwargs,
) -> type:
    """Per types.new_class"""
    resolved_bases = types.resolve_bases(bases)
    if resolved_bases is not bases:
        if '__orig_bases__' in namespace:
            raise TypeError((bases, resolved_bases))
        namespace['__orig_bases__'] = bases
    return super_meta.__new__(meta, name, resolved_bases, dict(namespace), **kwargs)


##


def deep_subclasses(cls: type, *, concrete_only: bool = False) -> ta.Iterator[type]:
    seen = set()
    todo = list(reversed(cls.__subclasses__()))
    while todo:
        cur = todo.pop()
        if cur in seen:
            continue
        seen.add(cur)
        if not (concrete_only and is_abstract_class(cur)):
            yield cur
        todo.extend(reversed(cur.__subclasses__()))


def build_mro_dict(
        instance_cls: type,
        owner_cls: type | None = None,
        *,
        bottom_up_key_order: bool = False,
) -> ta.Mapping[str, ta.Any]:
    if owner_cls is None:
        owner_cls = instance_cls

    mro = instance_cls.__mro__[-2::-1]
    try:
        pos = mro.index(owner_cls)
    except ValueError:
        raise TypeError(f'Owner class {owner_cls} not in mro of instance class {instance_cls}') from None

    dct: dict[str, ta.Any] = {}
    if not bottom_up_key_order:
        for cur_cls in mro[:pos + 1][::-1]:
            for k, v in cur_cls.__dict__.items():
                if k not in dct:
                    dct[k] = v

    else:
        for cur_cls in mro[:pos + 1]:
            dct.update(cur_cls.__dict__)

    return dct


##


class SimpleProxy(ta.Generic[T]):

    class Descriptor:

        def __init__(self, attr: str) -> None:
            super().__init__()
            self._attr = attr

        def __get__(self, instance, owner=None):
            if instance is None:
                return self
            return getattr(object.__getattribute__(instance, '__wrapped__'), self._attr)

        def __set__(self, instance, value):
            if instance is None:
                return self
            setattr(object.__getattribute__(instance, '__wrapped__'), self._attr, value)
            return None

        def __delete__(self, instance):
            if instance is None:
                return self
            delattr(object.__getattribute__(instance, '__wrapped__'), self._attr)
            return None

    __wrapped_attrs__: ta.ClassVar[ta.Iterable[str]] = ()

    def __init__(self, wrapped: T) -> None:
        super().__init__()
        object.__setattr__(self, '__wrapped__', wrapped)

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        for attr in cls.__wrapped_attrs__:
            setattr(cls, attr, SimpleProxy.Descriptor(attr))

    def __getattr__(self, item):
        return getattr(object.__getattribute__(self, '__wrapped__'), item)

    def __setattr__(self, name, value):
        setattr(object.__getattribute__(self, '__wrapped__'), name, value)

    def __delattr__(self, item):
        delattr(object.__getattribute__(self, '__wrapped__'), item)



from .... import lang
from .base import Keyword
from .base import StrKeyword


##


class MetadataKeyword(Keyword, lang.Abstract, lang.Sealed):
    pass


##


class Title(StrKeyword, MetadataKeyword, lang.Final, tag='title'):
    pass


class Description(StrKeyword, MetadataKeyword, lang.Final, tag='description'):
    pass



import json
import string
import typing as ta
import warnings

from .exceptions import EmptyExpressionError
from .exceptions import LexerError


class Lexer:
    START_IDENTIFIER: ta.AbstractSet[str] = set(string.ascii_letters + '_')
    VALID_IDENTIFIER: ta.AbstractSet[str] = set(string.ascii_letters + string.digits + '_')

    VALID_NUMBER: ta.AbstractSet[str] = set(string.digits)

    WHITESPACE: ta.AbstractSet[str] = set(' \t\n\r')

    SIMPLE_TOKENS: ta.Mapping[str, str] = {
        '.': 'dot',
        '*': 'star',
        ']': 'rbracket',
        ',': 'comma',
        ':': 'colon',
        '@': 'current',
        '(': 'lparen',
        ')': 'rparen',
        '{': 'lbrace',
        '}': 'rbrace',
        '+': 'plus',
        '%': 'modulo',
        '\u2212': 'minus',
        '\u00d7': 'multiply',
        '\u00f7': 'divide',
    }

    def __init__(self):
        self._enable_legacy_literals = False

    def tokenize(self, expression, options=None):
        if options is not None:
            self._enable_legacy_literals = options.enable_legacy_literals

        self._initialize_for_expression(expression)
        while self._current is not None:
            if self._current in self.SIMPLE_TOKENS:
                yield {
                    'type': self.SIMPLE_TOKENS[self._current],
                    'value': self._current,
                    'start': self._position,
                    'end': self._position + 1,
                }
                self._next()

            elif self._current in self.START_IDENTIFIER:
                start = self._position

                buff = self._current
                while self._next() in self.VALID_IDENTIFIER:
                    buff += self._current

                yield {
                    'type': 'unquoted_identifier',
                    'value': buff,
                    'start': start,
                    'end': start + len(buff),
                }

            elif self._current in self.WHITESPACE:
                self._next()

            elif self._current == '[':
                start = self._position

                next_char = self._next()
                if next_char == ']':
                    self._next()
                    yield {
                        'type': 'flatten',
                        'value': '[]',
                        'start': start,
                        'end': start + 2,
                    }

                elif next_char == '?':
                    self._next()
                    yield {
                        'type': 'filter',
                        'value': '[?',
                        'start': start,
                        'end': start + 2,
                    }

                else:
                    yield {
                        'type': 'lbracket',
                        'value': '[',
                        'start': start,
                        'end': start + 1,
                    }

            elif self._current == "'":
                yield self._consume_raw_string_literal()

            elif self._current == '|':
                yield self._match_or_else('|', 'or', 'pipe')

            elif self._current == '&':
                yield self._match_or_else('&', 'and', 'expref')

            elif self._current == '`':
                yield self._consume_literal()

            elif self._current in self.VALID_NUMBER:
                start = self._position

                buff = self._consume_number()
                yield {
                    'type': 'number',
                    'value': int(buff),
                    'start': start,
                    'end': start + len(buff),
                }

            elif self._current == '-':
                if not self._peek_is_next_digit():
                    self._next()
                    yield {
                        'type': 'minus',
                        'value': '-',
                        'start': self._position - 1,
                        'end': self._position,
                    }
                else:
                    # Negative number.
                    start = self._position
                    buff = self._consume_number()
                    if len(buff) > 1:
                        yield {
                            'type': 'number',
                            'value': int(buff),
                            'start': start,
                            'end': start + len(buff),
                        }
                    else:
                        raise LexerError(
                            lexer_position=start,
                            lexer_value=buff,
                            message=f"Unknown token '{buff}'")

            elif self._current == '/':
                self._next()
                if self._current == '/':
                    self._next()
                    yield {
                        'type': 'div',
                        'value': '//',
                        'start': self._position - 1,
                        'end': self._position,
                    }
                else:
                    yield {
                        'type': 'divide',
                        'value': '/',
                        'start': self._position,
                        'end': self._position + 1,
                    }

            elif self._current == '"':
                yield self._consume_quoted_identifier()

            elif self._current == '<':
                yield self._match_or_else('=', 'lte', 'lt')

            elif self._current == '>':
                yield self._match_or_else('=', 'gte', 'gt')

            elif self._current == '!':
                yield self._match_or_else('=', 'ne', 'not')

            elif self._current == '=':
                yield self._match_or_else('=', 'eq', 'assign')

            elif self._current == '$':
                if self._peek_may_be_valid_unquoted_identifier():
                    yield self._consume_variable()
                else:
                    yield {
                        'type': 'root',
                        'value': self._current,
                        'start': self._position,
                        'end': self._position + 1,
                    }
                    self._next()
            else:
                raise LexerError(
                    lexer_position=self._position,
                    lexer_value=self._current,
                    message=f'Unknown token {self._current})',
                )

        yield {
            'type': 'eof',
            'value': '',
            'start': self._length,
            'end': self._length,
        }

    def _consume_number(self):
        start = self._position  # noqa

        buff = self._current
        while self._next() in self.VALID_NUMBER:
            buff += self._current
        return buff

    def _consume_variable(self):
        start = self._position

        buff = self._current
        self._next()
        if self._current not in self.START_IDENTIFIER:
            raise LexerError(
                lexer_position=start,
                lexer_value=self._current,
                message=f'Invalid variable starting character {self._current}',
            )

        buff += self._current
        while self._next() in self.VALID_IDENTIFIER:
            buff += self._current

        return {
            'type': 'variable',
            'value': buff,
            'start': start,
            'end': start + len(buff),
        }

    def _peek_may_be_valid_unquoted_identifier(self):
        if (self._position == self._length - 1):
            return False
        else:
            nxt = self._chars[self._position + 1]
            return nxt in self.START_IDENTIFIER

    def _peek_is_next_digit(self):
        if (self._position == self._length - 1):
            return False
        else:
            nxt = self._chars[self._position + 1]
            return nxt in self.VALID_NUMBER

    def _initialize_for_expression(self, expression):
        if not expression:
            raise EmptyExpressionError
        self._position = 0
        self._expression = expression
        self._chars = list(self._expression)
        self._current = self._chars[self._position]
        self._length = len(self._expression)

    def _next(self):
        if self._position == self._length - 1:
            self._current = None
        else:
            self._position += 1
            self._current = self._chars[self._position]
        return self._current

    def _consume_until(self, delimiter):
        # Consume until the delimiter is reached, allowing for the delimiter to be escaped with "\".
        start = self._position

        buff = ''
        self._next()
        while self._current != delimiter:
            if self._current == '\\':
                buff += '\\'
                self._next()

            if self._current is None:
                # We're at the EOF.
                raise LexerError(
                    lexer_position=start,
                    lexer_value=self._expression[start:],
                    message=f'Unclosed {delimiter} delimiter',
                )

            buff += self._current
            self._next()

        # Skip the closing delimiter.
        self._next()
        return buff

    def _consume_literal(self):
        start = self._position

        token = self._consume_until('`')
        lexeme = token.replace('\\`', '`')
        parsed_json = None
        try:
            # Assume it is valid JSON and attempt to parse.
            parsed_json = json.loads(lexeme)
        except ValueError:
            error = LexerError(
                lexer_position=start,
                lexer_value=self._expression[start:],
                message=f'Bad token %s `{token}`',
            )

            if not self._enable_legacy_literals:
                raise error  # noqa

            try:
                # Invalid JSON values should be converted to quoted JSON strings during the JEP-12 deprecation period.
                parsed_json = json.loads('"%s"' % lexeme.lstrip())  # noqa
                warnings.warn('deprecated string literal syntax', DeprecationWarning)
            except ValueError:
                raise LexerError(  # noqa
                    lexer_position=start,
                    lexer_value=self._expression[start:],
                    message=f'Bad token {lexeme}',
                )

        token_len = self._position - start
        return {
            'type': 'literal',
            'value': parsed_json,
            'start': start,
            'end': token_len,
        }

    def _consume_quoted_identifier(self):
        start = self._position

        lexeme = '"' + self._consume_until('"') + '"'
        try:
            token_len = self._position - start
            return {
                'type': 'quoted_identifier',
                'value': json.loads(lexeme),
                'start': start,
                'end': token_len,
            }

        except ValueError as e:
            error_message = str(e).split(':')[0]
            raise LexerError(  # noqa
                lexer_position=start,
                lexer_value=lexeme,
                message=error_message,
            )

    def _consume_raw_string_literal(self):
        start = self._position

        lexeme = self._consume_until("'") \
            .replace("\\'", "'")  \
            .replace('\\\\', '\\')

        token_len = self._position - start
        return {
            'type': 'literal',
            'value': lexeme,
            'start': start,
            'end': token_len,
        }

    def _match_or_else(self, expected, match_type, else_type):
        start = self._position

        current = self._current
        next_char = self._next()
        if next_char == expected:
            self._next()
            return {
                'type': match_type,
                'value': current + next_char,
                'start': start,
                'end': start + 1,
            }

        return {
            'type': else_type,
            'value': current,
            'start': start,
            'end': start,
        }



import collections.abc
import contextlib
import dataclasses as dc
import keyword
import sys
import types
import typing as ta

from ... import check as check_
from .internals import PARAMS_ATTR
from .internals import Params
from .main import process_class
from .metadata import METADATA_ATTR
from .metadata import Metadata
from .params import FieldExtras
from .params import ParamsExtras


MISSING = dc.MISSING


def field(  # noqa
        default=MISSING,
        *,
        default_factory=MISSING,
        init=True,
        repr=True,  # noqa
        hash=None,  # noqa
        compare=True,
        metadata=None,
        kw_only=MISSING,

        derive: ta.Callable[..., ta.Any] | None = None,
        coerce: bool | ta.Callable[[ta.Any], ta.Any] | None = None,
        validate: ta.Callable[[ta.Any], bool] | None = None,
        check_type: bool | type | tuple[type | None, ...] | None = None,
        override: bool = False,
        repr_fn: ta.Callable[[ta.Any], str | None] | None = None,
        frozen: bool | None = None,
):  # -> dc.Field
    if default is not MISSING and default_factory is not MISSING:
        raise ValueError('cannot specify both default and default_factory')

    fx = FieldExtras(
        derive=derive,
        coerce=coerce,
        validate=validate,
        check_type=check_type,
        override=override,
        repr_fn=repr_fn,
        frozen=frozen,
    )

    md: ta.Mapping = {FieldExtras: fx}
    if metadata is not None:
        md = collections.ChainMap(md, check_.isinstance(metadata, collections.abc.Mapping))  # type: ignore

    return dc.Field(
        default,
        default_factory,  # noqa
        init,
        repr,
        hash,
        compare,
        types.MappingProxyType(md),
        kw_only,  # noqa
    )


def _strip_missing_values(d):
    return {k: v for k, v in d.items() if v is not MISSING}


def dataclass(  # noqa
        cls=None,
        /,
        *,
        init=True,
        repr=True,  # noqa
        eq=True,
        order=False,
        unsafe_hash=False,
        frozen=False,
        match_args=True,
        kw_only=False,
        slots=False,
        weakref_slot=False,

        metadata=None,

        reorder=MISSING,
        cache_hash=MISSING,
        generic_init=MISSING,
):
    def wrap(cls):
        pkw = dict(
            init=init,
            repr=repr,
            eq=eq,
            order=order,
            unsafe_hash=unsafe_hash,
            frozen=frozen,
            match_args=match_args,
            kw_only=kw_only,
            slots=slots,
            weakref_slot=weakref_slot,
        )

        dmd = cls.__dict__.get(METADATA_ATTR)

        epk = dict(dmd.get(_ExtraParamsKwargs, ()) if dmd is not None else ())
        epk.update(_strip_missing_values(dict(
            reorder=reorder,
            cache_hash=cache_hash,
            generic_init=generic_init,
        )))
        pex = ParamsExtras(**epk)

        mmd: dict = {
            ParamsExtras: pex,
        }

        md: Metadata = mmd
        cmds = []
        if metadata is not None:
            cmds.append(check_.isinstance(metadata, collections.abc.Mapping))
        if dmd is not None:
            cmds.append(dmd)
        if cmds:
            md = collections.ChainMap(md, *cmds)  # type: ignore

        setattr(cls, PARAMS_ATTR, Params(**pkw))
        setattr(cls, METADATA_ATTR, types.MappingProxyType(md))

        return process_class(cls)

    if cls is None:
        return wrap

    return wrap(cls)


def make_dataclass(  # noqa
        cls_name,
        fields,
        *,
        bases=(),
        namespace=None,
        init=True,
        repr=True,  # noqa
        eq=True,
        order=False,
        unsafe_hash=False,
        frozen=False,
        match_args=True,
        kw_only=False,
        slots=False,
        weakref_slot=False,
        module=None,

        reorder=MISSING,
        cache_hash=MISSING,
        generic_init=MISSING,
):
    if namespace is None:
        namespace = {}

    seen = set()
    annotations = {}
    defaults = {}
    for item in fields:
        if isinstance(item, str):
            name = item
            tp = 'typing.Any'
        elif len(item) == 2:
            name, tp, = item
        elif len(item) == 3:
            name, tp, spec = item
            defaults[name] = spec
        else:
            raise TypeError(f'Invalid field: {item!r}')
        if not isinstance(name, str) or not name.isidentifier():
            raise TypeError(f'Field names must be valid identifiers: {name!r}')
        if keyword.iskeyword(name):
            raise TypeError(f'Field names must not be keywords: {name!r}')
        if name in seen:
            raise TypeError(f'Field name duplicated: {name!r}')

        seen.add(name)
        annotations[name] = tp

    def exec_body_callback(ns):
        ns.update(namespace)
        ns.update(defaults)
        ns['__annotations__'] = annotations

    cls = types.new_class(cls_name, bases, {}, exec_body_callback)

    if module is None:
        try:
            module = sys._getframemodulename(1) or '__main__'  # type: ignore  # noqa
        except AttributeError:
            with contextlib.suppress(AttributeError, ValueError):
                module = sys._getframe(1).f_globals.get('__name__', '__main__')  # noqa
    if module is not None:
        cls.__module__ = module

    return dataclass(
        cls,
        init=init,
        repr=repr,
        eq=eq,
        order=order,
        unsafe_hash=unsafe_hash,
        frozen=frozen,
        match_args=match_args,
        kw_only=kw_only,
        slots=slots,
        weakref_slot=weakref_slot,

        reorder=reorder,
        cache_hash=cache_hash,
        generic_init=generic_init,
    )


class _ExtraParamsKwargs:
    pass


def extra_params(  # noqa
        *,
        reorder=MISSING,
        cache_hash=MISSING,
        generic_init=MISSING,
):
    def inner(cls):
        if PARAMS_ATTR in cls.__dict__:
            raise TypeError(cls)
        try:
            md = cls.__dict__[METADATA_ATTR]
        except KeyError:
            md = {}
            setattr(cls, METADATA_ATTR, md)
        if _ExtraParamsKwargs in md:
            raise TypeError(cls)

        md[_ExtraParamsKwargs] = _strip_missing_values(dict(
            reorder=reorder,
            cache_hash=cache_hash,
            generic_init=generic_init,
        ))

        return cls
    return inner



"""
TODO:
 - metaclass:
  - cleanup confer
 - descriptors - check_type/validators don't handle setters lol
 - deep_frozen?
 - field:
  - frozen
  - pickle/transient
  - mangled
  - doc
  - derive
  - check_type
 - class
  - strict_eq
  - allow_setattr
  - mangler
 - observable
 - c/py gen
 - iterable
 - proto/jsonschema gen
 - enums
 - nodal
 - embedding? forward kwargs in general? or only for replace?
"""



import typing as ta
import urllib.request

from .. import dataclasses as dc
from ..formats import json


@dc.dataclass(frozen=True)
class HubRepoInfo:
    repo: str
    tags: ta.Mapping[str, ta.Any]
    latest_manifests: ta.Mapping[str, ta.Any]


def get_hub_repo_info(
        repo: str,
        *,
        auth_url: str = 'https://auth.docker.io/',
        api_url: str = 'https://registry-1.docker.io/v2/',
) -> HubRepoInfo:
    """
    https://stackoverflow.com/a/39376254

    ==

    repo=library/nginx
    token=$(
        curl -s "https://auth.docker.io/token?service=registry.docker.io&scope=repository:${repo}:pull" \
        | jq -r '.token' \
    )
    curl -H "Authorization: Bearer $token" -s "https://registry-1.docker.io/v2/${repo}/tags/list" | jq
    curl \
        -H "Accept: application/vnd.docker.distribution.manifest.v2+json" \
        -H "Accept: application/vnd.docker.distribution.manifest.list.v2+json" \
        -H "Authorization: Bearer $token" \
        -s "https://registry-1.docker.io/v2/${repo}/manifests/latest" \
    | jq .
    """

    auth_url = auth_url.rstrip('/')
    api_url = api_url.rstrip('/')

    #

    def req_json(url: str, **kwargs: ta.Any) -> ta.Any:
        with urllib.request.urlopen(urllib.request.Request(url, **kwargs)) as resp:  # noqa
            return json.loads(resp.read().decode('utf-8'))

    #

    token_dct = req_json(f'{auth_url}/token?service=registry.docker.io&scope=repository:{repo}:pull')
    token = token_dct['token']

    req_hdrs = {'Authorization': f'Bearer {token}'}

    #

    tags_dct = req_json(
        f'{api_url}/{repo}/tags/list',
        headers=req_hdrs,
    )

    latest_mani_dct = req_json(
        f'{api_url}/{repo}/manifests/latest',
        headers={
            **req_hdrs,
            'Accept': 'application/vnd.docker.distribution.manifest.v2+json',
        },
    )

    return HubRepoInfo(
        repo,
        tags_dct,
        latest_mani_dct,
    )



from . import consts  # noqa

from .cookies import (  # noqa
    CookieTooBigError,
    dump_cookie,
    parse_cookie,
)

from .dates import (  # noqa
    http_date,
    parse_date,
)

from .encodings import (  # noqa
    latin1_decode,
    latin1_encode,
)

from .json import (  # noqa
    JSON_TAGGER,
    JsonTag,
    JsonTagger,
    json_dumps,
    json_loads,
)



import numbers
import operator
import typing as ta

from . import exceptions
from . import functions
from .scope import ScopedChainDict


def _equals(x, y):
    if _is_special_number_case(x, y):
        return False
    else:
        return x == y


def _is_special_number_case(x, y):
    # We need to special case comparing 0 or 1 to True/False.  While normally comparing any integer other than 0/1 to
    # True/False will always return False.  However 0/1 have this:
    # >>> 0 == True
    # False
    # >>> 0 == False
    # True
    # >>> 1 == True
    # True
    # >>> 1 == False
    # False
    #
    # Also need to consider that:
    # >>> 0 in [True, False]
    # True
    if _is_actual_number(x) and x in (0, 1):
        return isinstance(y, bool)

    elif _is_actual_number(y) and y in (0, 1):
        return isinstance(x, bool)

    else:
        return None


def _is_comparable(x):
    # The spec doesn't officially support string types yet, but enough people are relying on this behavior that it's
    # been added back.  This should eventually become part of the official spec.
    return _is_actual_number(x) or isinstance(x, str)


def _is_actual_number(x):
    # We need to handle python's quirkiness with booleans, specifically:
    #
    # >>> isinstance(False, int)
    # True
    # >>> isinstance(True, int)
    # True
    if isinstance(x, bool):
        return False
    return isinstance(x, numbers.Number)


class Options:
    """Options to control how a Jmespath function is evaluated."""

    def __init__(
            self,
            dict_cls=None,
            custom_functions=None,
            enable_legacy_literals=False,
    ):
        #: The class to use when creating a dict.  The interpreter may create dictionaries during the evaluation of a
        #  Jmespath expression.  For example, a multi-select hash will create a dictionary.  By default we use a dict()
        #  type. You can set this value to change what dict type is used. The most common reason you would change this
        #  is if you want to set a collections.OrderedDict so that you can have predictable key ordering.
        self.dict_cls = dict_cls
        self.custom_functions = custom_functions

        #: The flag to enable pre-JEP-12 literal compatibility.
        #  JEP-12 deprecates `foo` -> "foo" syntax.
        #  Valid expressions MUST use: `"foo"` -> "foo"
        #  Setting this flag to `True` enables support for legacy syntax.
        self.enable_legacy_literals = enable_legacy_literals


class _Expression:
    def __init__(self, expression, interpreter):
        self.expression = expression
        self.interpreter = interpreter

    def visit(self, node, *args, **kwargs):
        return self.interpreter.visit(node, *args, **kwargs)


class Visitor:
    def __init__(self):
        self._method_cache = {}

    def visit(self, node, *args, **kwargs):
        node_type = node['type']
        method = self._method_cache.get(node_type)
        if method is None:
            method = getattr(self, f'visit_{node["type"]}', self.default_visit)
            self._method_cache[node_type] = method
        return method(node, *args, **kwargs)  # type: ignore

    def default_visit(self, node, *args, **kwargs):
        raise NotImplementedError('default_visit')


class TreeInterpreter(Visitor):
    COMPARATOR_FUNC: ta.Mapping[str, ta.Callable] = {
        'eq': _equals,
        'ne': lambda x, y: not _equals(x, y),
        'lt': operator.lt,
        'gt': operator.gt,
        'lte': operator.le,
        'gte': operator.ge,
    }

    _EQUALITY_OPS: ta.Sequence[str] = ['eq', 'ne']

    _ARITHMETIC_UNARY_FUNC: ta.Mapping[str, ta.Callable] = {
        'minus': operator.neg,
        'plus': lambda x: x,
    }

    _ARITHMETIC_FUNC: ta.Mapping[str, ta.Callable] = {
        'div': operator.floordiv,
        'divide': operator.truediv,
        'minus': operator.sub,
        'modulo': operator.mod,
        'multiply': operator.mul,
        'plus': operator.add,
    }

    MAP_TYPE = dict

    def __init__(self, options=None):
        super().__init__()

        self._dict_cls = self.MAP_TYPE

        if options is None:
            options = Options()
        self._options = options

        if options.dict_cls is not None:
            self._dict_cls = self._options.dict_cls

        if options.custom_functions is not None:
            self._functions = self._options.custom_functions
        else:
            self._functions = functions.Functions()

        self._root = None
        self._scope = ScopedChainDict()

    def default_visit(self, node, *args, **kwargs):
        raise NotImplementedError(node['type'])

    def evaluate(self, ast, root):
        self._root = root
        return self.visit(ast, root)

    def visit_subexpression(self, node, value):
        result = value
        for child in node['children']:
            result = self.visit(child, result)
            if result is None:
                return None
        return result

    def visit_field(self, node, value):
        try:
            return value.get(node['value'])
        except AttributeError:
            return None

    def visit_comparator(self, node, value):
        # Common case: comparator is == or !=
        comparator_func = self.COMPARATOR_FUNC[node['value']]
        if node['value'] in self._EQUALITY_OPS:
            return comparator_func(
                self.visit(node['children'][0], value),
                self.visit(node['children'][1], value),
            )

        else:
            # Ordering operators are only valid for numbers. Evaluating any other type with a comparison operator will
            # yield a None value.
            left = self.visit(node['children'][0], value)
            right = self.visit(node['children'][1], value)
            # num_types = (int, float)
            if not (_is_comparable(left) and _is_comparable(right)):
                return None
            return comparator_func(left, right)

    def visit_arithmetic_unary(self, node, value):
        operation = self._ARITHMETIC_UNARY_FUNC[node['value']]
        return operation(
            self.visit(node['children'][0], value),
        )

    def visit_arithmetic(self, node, value):
        operation = self._ARITHMETIC_FUNC[node['value']]
        return operation(
            self.visit(node['children'][0], value),
            self.visit(node['children'][1], value),
        )

    def visit_current(self, node, value):
        return value

    def visit_root(self, node, value):
        return self._root

    def visit_expref(self, node, value):
        return _Expression(node['children'][0], self)

    def visit_function_expression(self, node, value):
        resolved_args = []
        for child in node['children']:
            current = self.visit(child, value)
            resolved_args.append(current)

        return self._functions.call_function(node['value'], resolved_args)

    def visit_filter_projection(self, node, value):
        base = self.visit(node['children'][0], value)
        if not isinstance(base, list):
            return None

        comparator_node = node['children'][2]
        collected = []
        for element in base:
            if self._is_true(self.visit(comparator_node, element)):
                current = self.visit(node['children'][1], element)
                if current is not None:
                    collected.append(current)

        return collected

    def visit_flatten(self, node, value):
        base = self.visit(node['children'][0], value)
        if not isinstance(base, list):
            # Can't flatten the object if it's not a list.
            return None

        merged_list = []
        for element in base:
            if isinstance(element, list):
                merged_list.extend(element)
            else:
                merged_list.append(element)

        return merged_list

    def visit_identity(self, node, value):
        return value

    def visit_index(self, node, value):
        # Even though we can index strings, we don't want to support that.
        if not isinstance(value, list):
            return None

        try:
            return value[node['value']]
        except IndexError:
            return None

    def visit_index_expression(self, node, value):
        result = value
        for child in node['children']:
            result = self.visit(child, result)

        return result

    def visit_slice(self, node, value):
        if isinstance(value, str):
            start = node['children'][0]
            end = node['children'][1]
            step = node['children'][2]
            return value[start:end:step]

        if not isinstance(value, list):
            return None

        s = slice(*node['children'])
        return value[s]

    def visit_key_val_pair(self, node, value):
        return self.visit(node['children'][0], value)

    def visit_literal(self, node, value):
        return node['value']

    def visit_multi_select_dict(self, node, value):
        collected = self._dict_cls()
        for child in node['children']:
            collected[child['value']] = self.visit(child, value)

        return collected

    def visit_multi_select_list(self, node, value):
        collected = []
        for child in node['children']:
            collected.append(self.visit(child, value))

        return collected

    def visit_or_expression(self, node, value):
        matched = self.visit(node['children'][0], value)

        if self._is_false(matched):
            matched = self.visit(node['children'][1], value)

        return matched

    def visit_and_expression(self, node, value):
        matched = self.visit(node['children'][0], value)

        if self._is_false(matched):
            return matched

        return self.visit(node['children'][1], value)

    def visit_not_expression(self, node, value):
        original_result = self.visit(node['children'][0], value)

        if _is_actual_number(original_result) and original_result == 0:
            # Special case for 0, !0 should be false, not true. 0 is not a special cased integer in jmespath.
            return False

        return not original_result

    def visit_pipe(self, node, value):
        result = value
        for child in node['children']:
            result = self.visit(child, result)
        return result

    def visit_projection(self, node, value):
        base = self.visit(node['children'][0], value)

        allow_string = False
        first_child = node['children'][0]
        if first_child['type'] == 'index_expression':
            nested_children = first_child['children']
            if len(nested_children) > 1 and nested_children[1]['type'] == 'slice':
                allow_string = True

        if isinstance(base, str) and allow_string:
            # projections are really sub-expressions in disguise evaluate the rhs when lhs is a sliced string
            return self.visit(node['children'][1], base)

        if not isinstance(base, list):
            return None
        collected = []
        for element in base:
            current = self.visit(node['children'][1], element)
            if current is not None:
                collected.append(current)

        return collected

    def visit_let_expression(self, node, value):
        *bindings, expr = node['children']
        scope = {}
        for assign in bindings:
            scope.update(self.visit(assign, value))
        self._scope.push_scope(scope)
        result = self.visit(expr, value)
        self._scope.pop_scope()
        return result

    def visit_assign(self, node, value):
        name = node['value']
        value = self.visit(node['children'][0], value)
        return {name: value}

    def visit_variable_ref(self, node, value):
        try:
            return self._scope[node['value']]
        except KeyError:
            raise exceptions.UndefinedVariableError(node['value'])  # noqa

    def visit_value_projection(self, node, value):
        base = self.visit(node['children'][0], value)
        try:
            base = base.values()
        except AttributeError:
            return None

        collected = []
        for element in base:
            current = self.visit(node['children'][1], element)
            if current is not None:
                collected.append(current)

        return collected

    def _is_false(self, value):
        # This looks weird, but we're explicitly using equality checks because the truth/false values are different
        # between python and jmespath.
        return (value == '' or value == [] or value == {} or value is None or value is False)  # noqa

    def _is_true(self, value):
        return not self._is_false(value)


class GraphvizVisitor(Visitor):
    def __init__(self):
        super().__init__()
        self._lines = []
        self._count = 1

    def visit(self, node, *args, **kwargs):
        self._lines.append('digraph AST {')
        current = f"{node['type']}{self._count}"
        self._count += 1
        self._visit(node, current)
        self._lines.append('}')
        return '\n'.join(self._lines)

    def _visit(self, node, current):
        self._lines.append('%s [label="%s(%s)"]' % (current, node['type'], node.get('value', '')))  # noqa
        for child in node.get('children', []):
            child_name = f"{child['type']}{self._count}"
            self._count += 1
            self._lines.append(f'  {current} -> {child_name}')
            self._visit(child, child_name)



from .asyncs import (  # noqa
    SyncableIterable,
    async_list,
    sync_await,
    sync_list,
    syncable_iterable,
)

from .bridge import (  # noqa
    a_to_s,
    is_in_bridge,
    s_to_a,
    s_to_a_await,
    trivial_a_to_s,
    trivial_s_to_a,
)

from .flavors import (  # noqa
    ContextManagerAdapter,
    Flavor,
    adapt,
    adapt_context,
    from_anyio,
    from_anyio_context,
    from_asyncio,
    from_asyncio_context,
    from_trio,
    from_trio_context,
    get_flavor,
    mark_anyio,
    mark_asyncio,
    mark_flavor,
    mark_trio,
    with_adapter_loop,
)



"""
TODO:
 - auto drain_asyncio

==

_pytest:
https://github.com/pytest-dev/pytest/blob/ef9b8f9d748b6f50eab5d43e32d93008f7880899/src/_pytest/python.py#L155
"async def function and no async plugin installed (see warnings)"

pytest_trio:
https://github.com/python-trio/pytest-trio/blob/f03160aa1dd355a12d39fa21f4aee4e1239efea3/pytest_trio/plugin.py
 - pytest_addoption
 - pytest_configure
 - pytest_collection_modifyitems
 - pytest_runtest_call @(hookwrapper=True)
 - pytest_fixture_setup

pytest_asyncio:
https://github.com/pytest-dev/pytest-asyncio/blob/f45aa18cf3eeeb94075de16a1d797858facab863/pytest_asyncio/plugin.py
 - pytest_addoption
 - pytest_configure
 - pytest_pycollect_makeitem_preprocess_async_fixtures @(specname="pytest_pycollect_makeitem", tryfirst=True)
 - pytest_pycollect_makeitem_convert_async_functions_to_subclass @(specname="pytest_pycollect_makeitem", hookwrapper=True)
 - pytest_generate_tests @(tryfirst=True)
 - pytest_runtest_setup(item: pytest.Item) -> None:
 - pytest_pyfunc_call @(tryfirst=True, hookwrapper=True)
 - pytest_collectstart @()
 - pytest_report_header @(tryfirst=True)
 - pytest_fixture_setup @(hookwrapper=True)

anyio.pytest_plugin:
https://github.com/agronholm/anyio/blob/8907964926a24461840eee0925d3f355e729f15d/src/anyio/pytest_plugin.py
 - pytest_configure
 - pytest_pycollect_makeitem @(tryfirst=True)
 - pytest_pyfunc_call @(tryfirst=True)
 - pytest_fixture_setup

"""  # noqa
import functools
import inspect
import sys
import typing as ta
import warnings

import pytest

from .... import lang
from ....asyncs import trio_asyncio as trai
from ....diag import pydevd as pdu
from ._registry import register


if ta.TYPE_CHECKING:
    import trio_asyncio
else:
    trio_asyncio = lang.proxy_import('trio_asyncio')


ASYNCS_MARK = 'asyncs'

KNOWN_BACKENDS = (
    'asyncio',
    'trio',
    'trio_asyncio',
)

PARAM_NAME = '__async_backend'


def iscoroutinefunction(func: ta.Any) -> bool:
    return inspect.iscoroutinefunction(func) or getattr(func, '_is_coroutine', False)


def is_async_function(func: ta.Any) -> bool:
    return iscoroutinefunction(func) or inspect.isasyncgenfunction(func)


@register
class AsyncsPlugin:
    ASYNC_BACKENDS: ta.ClassVar[ta.Sequence[str]] = [
        *[s for s in KNOWN_BACKENDS if lang.can_import(s)],
    ]

    def pytest_cmdline_main(self, config):
        if (aio_plugin := sys.modules.get('pytest_asyncio.plugin')):
            # warnings.filterwarnings is clobbered by pytest using warnings.catch_warnings
            def aio_plugin_warn(message, *args, **kwargs):
                if (
                        isinstance(message, pytest.PytestDeprecationWarning) and
                        message.args[0].startswith('The configuration option "asyncio_default_fixture_loop_scope" is unset.')  # noqa
                ):
                    return
                warnings.warn(message, *args, **kwargs)

            aio_plugin.warnings = lang.proxy_import('warnings')  # type: ignore
            aio_plugin.warnings.warn = aio_plugin_warn  # type: ignore

    def pytest_configure(self, config):
        config.addinivalue_line('markers', f'{ASYNCS_MARK}: marks for all async backends')
        config.addinivalue_line('markers', 'trio_asyncio: marks for trio_asyncio backend')

    def pytest_generate_tests(self, metafunc):
        if (m := metafunc.definition.get_closest_marker(ASYNCS_MARK)) is not None:
            if m.args:
                bes = m.args
            else:
                bes = self.ASYNC_BACKENDS
        elif metafunc.definition.get_closest_marker('trio_asyncio') is not None:
            bes = ['trio_asyncio']
        else:
            return

        if 'trio_asyncio' in bes:
            # NOTE: Importing it here is apparently necessary to get its patching working - otherwise fails later with
            # `no running event loop` in anyio._backends._asyncio and such.
            import trio_asyncio  # noqa

        if pdu.is_present():
            pdu.patch_for_trio_asyncio()

        metafunc.fixturenames.append(PARAM_NAME)
        metafunc.parametrize(PARAM_NAME, bes)

        for c in metafunc._calls:  # noqa
            be = c.params[PARAM_NAME]
            if be == 'trio_asyncio':
                c.marks.extend([
                    pytest.mark.trio.mark,
                    pytest.mark.trio_asyncio.mark,
                ])

            else:
                c.marks.append(getattr(pytest.mark, be).mark)

            if pdu.is_present():
                c.marks.append(pytest.mark.drain_asyncio.mark)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_call(self, item):
        bes = [be for be in self.ASYNC_BACKENDS if item.get_closest_marker(be) is not None]
        if len(bes) > 1 and set(bes) != {'trio', 'trio_asyncio'}:
            raise Exception(f'{item.nodeid}: multiple async backends specified: {bes}')
        elif is_async_function(item.obj) and not bes:
            raise Exception(f'{item.nodeid}: async def function and no async plugin specified')

        if 'trio_asyncio' in bes:
            obj = item.obj

            @functools.wraps(obj)
            @trai.with_trio_asyncio_loop(wait=True)
            async def run(*args, **kwargs):
                await trio_asyncio.aio_as_trio(obj)(*args, **kwargs)

            item.obj = run

        yield
